function [x,er,y]=newtonsolve(fname,x,opt,varargin)
% NewtonSolve - Numerical solution of nonlinear equations
%               using Newton or Quasi-Newton methods.
%
% [x,er,y]=NewtonSolve(fname,x,opt,z1,z2,...zn)
%
% Input:  fname - name of the m-file implementing f(x)
%         x - initial guess
%         opt - vector of options
%               opt(1): TolX in stopping rule (>eps)
%               opt(1): TolY in convergence test (>eps)
%               opt(3): Maximum number of iterations (>1)
%               opt(4): Algorithm
%                       1: Newton
%                       2: Broyden unfactored (with restarting)
%                       3: Broyden factored (with restarting)
%               (if opt=[] or missing then the following default values are used:
%               TolX=sqrt(esp), TolY=sqrt(esp), MaxIer=500, Alg=1)
%         z1,z2,...,zn - optional arguments passed to fname
%
% Stopping rules: (1) norm(y)<=TolY (convergence to a solution)
%                 (2) norm(step)<=TolX*(1+norm(x)) (stagnation: the algorithm reports
%                     success iif norm(y)>TolY*(1+norm(y0)))
%                 (3) the maximum number of iterations has been reached
%
% Output: x - solution of f(x)=0
%         er - error code
%              0: OK, Convergence to a solution
%              1: Convergence to a point that is not a solution
%              2: Maximum number of iterations reached
%         y - value of f(x) at x
%
% External calls: Jacob1.m (one-sided finite diff. Jacobian) using Alg=1
%
% Author: Marco Maffezzoli. Ver. 3.1.2, 11/2012.
%
% This program is provided "as-is," without any express or implied
% warranty. Please, handle it with care!
%

if nargin<=1
    error('Sorry, I need at least two input arguments!')
elseif nargin==2||isempty(opt)
    tolx=sqrt(eps);
    toly=sqrt(eps);
    maxeval=500;
    alg=1;
else
    tolx=max(opt(1),eps);
    toly=max(opt(2),eps);
    maxeval=max(round(opt(3)),1);
    alg=round(opt(4));
end
shx=size(x);
objf=str2func(fname);
y=objf(x,varargin{:});
if numel(x)~=numel(y)
    error('The system is not square!')
end
ny=norm(y(:));
ny0=ny;
J=jacob1(objf,x,y,varargin{:});
switch alg
    case 1
        s=-J\y(:);
    case 2
        Jinv=J\eye(numel(x));
        s=-Jinv*y(:);
    otherwise
        [Q,R]=qr(J);
        s=-R\(Q.'*y(:));
end
for k=1:maxeval
    if norm(s)<=tolx*(1+norm(x(:)))
        if ny<=toly*(1+ny0)
            er=0;
        else
            er=1;
        end
        return
    end
    x=reshape(x(:)+s,shx);
    y=objf(x,varargin{:});
    ny1=norm(y(:));
    if (alg~=1)&&(ny1/ny<0.5)
        if alg==2 % Quasi-Newton: Broyden unfactored
            z=-Jinv*y(:);
            Jinv=Jinv+(z*s.'*Jinv)/(s.'*(s-z));
            s=-Jinv*y(:);
        else % Quasi-Newton: Broyden factored
            [Q,R]=qrupdate(Q,R,y(:)/(s.'*s),s);
            s=-R\(Q.'*y(:));
        end
    else % Newton
        J=jacob1(objf,x,y,varargin{:});
        s=-J\y(:);
    end
    ny=ny1;
end
er=2;

end